from Crypto.PublicKey import RSA

from sgid_client.error import Errors


def convert_to_pkcs8(private_key: str) -> str:
    """Converts a private key in PKCS1 format to PKCS8.

    Args:
        private_key (str): Private key as a string.

    Raises:
        Exception: if private key is invalid.

    Returns:
        str: Private key in PKCS8 format.
    """
    try:
        imported = RSA.import_key(extern_key=private_key)
        return imported.export_key(pkcs=8).decode("ascii")
    except Exception as exc:
        raise Exception(Errors.PRIVATE_KEY_IMPORT) from exc



from typing import NamedTuple
from datetime import datetime

from requests import Response


class SgidClientError(NamedTuple):
    MISSING_REDIRECT_URI: str
    TOKEN_ENDPOINT_FAILED: str
    ID_TOKEN_MALFORMED: str
    ID_TOKEN_HEADER_PAYLOAD_MALFORMED: str
    ID_TOKEN_WRONG_SIGNING_ALG: str
    ID_TOKEN_MISSING_KEYS: str
    ID_TOKEN_ISS_MISMATCH: str
    ID_TOKEN_IAT_INVALID: str
    ID_TOKEN_EXP_INVALID: str
    ID_TOKEN_EXPIRED: str
    ID_TOKEN_AUD_MISMATCH: str
    ID_TOKEN_NONCE_MISMATCH: str
    ID_TOKEN_SUB_INVALID: str
    ID_TOKEN_SIGNATURE_INVALID: str
    ACCESS_TOKEN_INVALID: str
    USERINFO_ENDPOINT_FAILED: str
    JWKS_ENDPOINT_FAILED: str
    PRIVATE_KEY_IMPORT: str
    USERINFO_SUB_MISMATCH: str
    USERINFO_BLOCK_KEY_DECRYPT_FAILED: str
    USERINFO_DATA_DECRYPT_FAILED: str
    CODE_VERIFIER_LENGTH_ERROR: str
    PKCE_PAIR_LENGTH_ERROR: str


Errors = SgidClientError(
    MISSING_REDIRECT_URI="No redirect URI registered with this client. You must either specify a valid redirect URI in the SgidClient constructor, or pass it to the authorization_url and callback functions.",
    TOKEN_ENDPOINT_FAILED="sgID responded with an error at the token endpoint",
    ID_TOKEN_MALFORMED="sgID server returned a malformed ID token which did not contain the expected components (header, payload, signature)",
    ID_TOKEN_HEADER_PAYLOAD_MALFORMED="ID token header or payload was malformed",
    ID_TOKEN_WRONG_SIGNING_ALG="Unexpected signing algorithm used for ID token",
    ID_TOKEN_MISSING_KEYS="ID token payload did not contain the required keys",
    ID_TOKEN_ISS_MISMATCH="ID token 'iss' did not match expected value",
    ID_TOKEN_IAT_INVALID="ID token 'iat' is invalid",
    ID_TOKEN_EXP_INVALID="ID token 'exp' is invalid",
    ID_TOKEN_EXPIRED="ID token is expired",
    ID_TOKEN_AUD_MISMATCH="ID token 'aud' did not match client ID",
    ID_TOKEN_NONCE_MISMATCH="ID token 'nonce' did not match the nonce passed to the callback function",
    ID_TOKEN_SUB_INVALID="ID token 'sub' is invalid",
    ID_TOKEN_SIGNATURE_INVALID="ID token signature is invalid",
    ACCESS_TOKEN_INVALID="sgID token endpoint did not return a valid access token. Expected a non-empty string.",
    USERINFO_ENDPOINT_FAILED="sgID responded with an error at the userinfo endpoint",
    JWKS_ENDPOINT_FAILED="sgID responded with an error at the jwks endpoint",
    USERINFO_SUB_MISMATCH="Sub returned by sgID did not match the sub passed to the userinfo method. Check that you passed the correct sub to the userinfo method.",
    PRIVATE_KEY_IMPORT="Failed to import private key. Check that privateKey is a valid PKCS1 or PKCS8 key.",
    USERINFO_BLOCK_KEY_DECRYPT_FAILED="Decryption of block key failed. Check that you passed the correct private key to the SgidClient constructor.",
    USERINFO_DATA_DECRYPT_FAILED="Decryption of data failed. Check that you passed the correct private key to the SgidClient constructor.",
    CODE_VERIFIER_LENGTH_ERROR="Code verifier should have a minimum length of 43 and a maximum length of 128",
    PKCE_PAIR_LENGTH_ERROR="generate_pkce_pair should receive a minimum length of 43 and a maximum length of 128",
)


def get_network_error_message(message: str, status: int, body: str) -> str:
    return f"{message}\nResponse status: {status}\nResponse body: {body}"


def get_expected_vs_received_error_message(message: str, expected, received) -> str:
    return f"{message}. Expected {expected}, received {received}."


def get_expiry_error_message(message: str, expired_at: float) -> str:
    curr_time = datetime.now()
    return f"{message}. Current timestamp is {curr_time.strftime('%Y/%m/%d, %H:%M:%S')}, expiry was at {datetime.fromtimestamp(expired_at).strftime('%Y/%m/%d, %H:%M:%S')}"


def get_www_authenticate_error_message(message: str, res: Response) -> str:
    www_authenticate_header = res.headers.get("www-authenticate", None)
    if www_authenticate_header is None:
        return get_network_error_message(
            message=message, status=res.status_code, body=res.text
        )

    if www_authenticate_header.startswith("error="):
        return f"{message}\nResponse status: {res.status_code}\nError message: {www_authenticate_header[len('error='):]}"
    return f"{message}\nResponse status: {res.status_code}\nError message: {www_authenticate_header}"



from jwcrypto import jwk, jwe
from .error import Errors


def decrypt_data(encrypted_key: str, encrypted_data: dict, private_key: str):
    try:
        # Load private_key
        private_key_jwk = jwk.JWK.from_pem(private_key.encode("ascii"))
    except Exception as exc:
        raise Exception(Errors.PRIVATE_KEY_IMPORT) from exc

    try:
        jwe_key = jwe.JWE()
        # Decrypt encrypted_key to get block_key
        jwe_key.deserialize(encrypted_key, key=private_key_jwk)
    except Exception as exc:
        raise Exception(Errors.USERINFO_BLOCK_KEY_DECRYPT_FAILED) from exc

    block_key_json = jwe_key.payload
    try:
        # Load block_key
        block_key = jwk.JWK.from_json(block_key_json.decode("ascii").replace("'", '"'))
        jwe_data = jwe.JWE()

        # Initialise dict
        data_dict = {}

        for field in encrypted_data:
            # Decrypt encrypted_data[field] to get actual_data
            jwe_data.deserialize(encrypted_data[field], key=block_key)
            data_dict[field] = jwe_data.payload.decode("ascii")
    except Exception as exc:
        raise Exception(Errors.USERINFO_DATA_DECRYPT_FAILED) from exc

    return data_dict



import requests
from jwcrypto import jwk, jwt as jwt_lib
from .error import Errors, get_network_error_message


class IdTokenVerifier:
    def __init__(self, jwks_uri: str) -> None:
        self.jwks_cache = None
        self.jwks_uri = jwks_uri

    def _fetch_jwks_from_uri(self):
        res = requests.get(self.jwks_uri)
        if res.status_code != 200:
            error_message = get_network_error_message(
                message=Errors.JWKS_ENDPOINT_FAILED,
                status=res.status_code,
                body=res.text,
            )
            raise Exception(error_message)
        return jwk.JWKSet.from_json(res.text)

    def verify_jwt(self, jwt: str):
        if self.jwks_cache is None:
            self.jwks_cache = self._fetch_jwks_from_uri()
        try:
            jwt_lib.JWT(key=self.jwks_cache, jwt=jwt)
        except:
            # Try again in case jwks has refreshed
            self.jwks_cache = self._fetch_jwks_from_uri()
            try:
                jwt_lib.JWT(key=self.jwks_cache, jwt=jwt)
            except:
                raise Exception(Errors.ID_TOKEN_SIGNATURE_INVALID)



import secrets
from typing import NamedTuple
from urllib.parse import urlparse, urlencode
import requests
from .validation import validate_access_token, validate_id_token
from .IdTokenVerifier import IdTokenVerifier
from .decrypt_data import decrypt_data
from .error import Errors, get_network_error_message, get_www_authenticate_error_message
from .util import convert_to_pkcs8

API_VERSION = 2
SGID_RESPONSE_TYPE = "code"
SGID_CODE_CHALLENGE_METHOD = "S256"
SGID_GRANT_TYPE = "authorization_code"


class AuthorizationUrlReturn(NamedTuple):
    url: str
    nonce: str | None


class CallbackReturn(NamedTuple):
    sub: str
    access_token: str


class UserInfoReturn(NamedTuple):
    sub: str
    data: dict


class SgidClient:
    """Class which allows you to interact with the sgID API."""

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        private_key: str,
        redirect_uri: str | None = None,
        hostname: str = "https://api.id.gov.sg",
    ):
        """Initialises an SgidClient instance.

        Args:
            client_id (str): Client ID provided during client registration.

            client_secret (str): Client secret provided during client registration.

            private_key (str): Client private key provided during client registration.

            redirect_uri (str | None, optional): Redirection URI for user to return to your
            application after login. If not provided in the constructor, this must
            be provided to the authorization_url and callback functions. Defaults to None.

            hostname (str, optional): Hostname of OpenID provider (sgID). Defaults to "https://api.id.gov.sg".

        Raises:
            Exception: if private key is invalid.
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.private_key = convert_to_pkcs8(private_key)
        self.redirect_uri = redirect_uri
        self.issuer = f"{urlparse(hostname).geturl()}/v{API_VERSION}"
        self.verifier = IdTokenVerifier(jwks_uri=f"{self.issuer}/.well-known/jwks.json")

    def authorization_url(
        self,
        code_challenge: str,
        state: str | None = None,
        redirect_uri: str | None = None,
        scope: str | list[str] = "openid myinfo.name",
        nonce: str | None = secrets.token_urlsafe(32),
    ) -> AuthorizationUrlReturn:
        """Generates authorization url to redirect end-user to sgID login page.

        Args:
            code_challenge (str): The code challenge generated from generate_pkce_pair().

            state (str | None, optional): A string which will be passed back to your application once
            the end-user logs in. You can also use this to track per-request state. Defaults to None.

            redirect_uri (str | None, optional): The redirect URI used in the authorization
            request. If this param is provided, it will be used instead of the redirect
            URI provided in the SgidClient constructor. If not provided in the constructor,
            the redirect URI must be provided here. Defaults to None.

            scope (str | list[str], optional): "openid" must be provided as a
            scope. Defaults to "openid myinfo.name".

            nonce (str | None, optional): Unique nonce for this request. If this param is
            not provided, a nonce is generated and returned. To prevent this behaviour,
            specify None for this param. Defaults to `secrets.token_urlsafe(32)`.

        Raises:
            Exception: if redirect URI is provided in neither the constructor nor this function.

        Returns:
            AuthorizationUrlReturn: authorization URL and nonce.
        """
        if redirect_uri is None and self.redirect_uri is None:
            raise Exception(Errors.MISSING_REDIRECT_URI)
        params = {
            "client_id": self.client_id,
            "scope": " ".join(scope) if isinstance(scope, list) else scope,
            "redirect_uri": self.redirect_uri if redirect_uri is None else redirect_uri,
            "response_type": SGID_RESPONSE_TYPE,
            "code_challenge": code_challenge,
            "code_challenge_method": SGID_CODE_CHALLENGE_METHOD,
        }
        if state is not None:
            params["state"] = state
        if nonce is not None:
            params["nonce"] = nonce
        auth_url = f"{self.issuer}/oauth/authorize?{urlencode(params)}"
        return AuthorizationUrlReturn(url=auth_url, nonce=nonce)

    def callback(
        self,
        code: str,
        code_verifier: str,
        nonce: str | None = None,
        redirect_uri: str | None = None,
    ) -> CallbackReturn:
        """Exchanges authorization code for access token.

        Args:
            code (str): The authorization code received from the authorization server.

            code_verifier (str): The code verifier corresponding to the code challenge
            that was passed to `authorization_url` for this request.

            nonce (str | None, optional): Nonce passed to `authorization_url` for this
            request. Specify None if no nonce was passed to `authorization_url`.
            Defaults to None.

            redirect_uri (str | None, optional): The redirect URI used in the
            authorization request. If not specified, defaults to the one passed
            to the SgidClient constructor.

        Raises:
            Exception: if call to token endpoint fails.
            Exception: if call to JWKS endpoint fails.
            Exception: if ID token validation fails.
            Exception: if access token validation fails.

        Returns:
            CallbackReturn: The sub (subject identifier claim) of the user and
            access token. The subject identifier claim is the end-user's unique ID.
        """
        url = f"{self.issuer}/oauth/token"
        data = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "redirect_uri": self.redirect_uri if redirect_uri is None else redirect_uri,
            "code": code,
            "grant_type": SGID_GRANT_TYPE,
            "code_verifier": code_verifier,
        }
        res = requests.post(url, data)
        if res.status_code != 200:
            error_message = get_network_error_message(
                message=Errors.TOKEN_ENDPOINT_FAILED,
                status=res.status_code,
                body=res.text,
            )
            raise Exception(error_message)
        res_body = res.json()
        id_token: str = res_body["id_token"]
        access_token: str = res_body["access_token"]
        sub = validate_id_token(
            id_token=id_token,
            issuer=self.issuer,
            client_id=self.client_id,
            nonce=nonce,
            verifier=self.verifier,
        )
        validate_access_token(access_token=access_token)
        return CallbackReturn(sub=sub, access_token=access_token)

    def userinfo(self, sub: str, access_token: str) -> UserInfoReturn:
        """Retrieves verified user info and decrypts it with your private key.

        Args:
            sub (str): The sub returned from the `callback` function.

            access_token (str): The access token returned from the `callback` function.

        Raises:
            Exception: if call to userinfo endpoint fails.
            Exception: if sub returned from userinfo endpoint does not match
            sub passed to this function.
            Exception: if decryption fails.

        Returns:
            UserInfoReturn: The sub of the end-user and the end-user's verified data.
            The subreturned is the same as the one passed in the params.
        """
        url = f"{self.issuer}/oauth/userinfo"
        headers = {
            "Authorization": f"Bearer {access_token}",
        }
        res = requests.get(url, headers=headers)
        if res.status_code != 200:
            error_message = get_www_authenticate_error_message(
                message=Errors.USERINFO_ENDPOINT_FAILED, res=res
            )
            raise Exception(error_message)
        res_body = res.json()
        if res_body["sub"] != sub:
            raise Exception(Errors.USERINFO_SUB_MISMATCH)
        decrypted_data = decrypt_data(
            encrypted_key=res_body["key"],
            encrypted_data=res_body["data"],
            private_key=self.private_key,
        )
        return UserInfoReturn(sub=res_body["sub"], data=decrypted_data)



from datetime import datetime
from sgid_client import IdTokenVerifier
import json
from base64 import b64decode
from urllib.parse import unquote
from .error import (
    Errors,
    get_expected_vs_received_error_message,
    get_expiry_error_message,
)


def validate_access_token(access_token: str):
    if type(access_token) is not str or access_token == "":
        raise Exception(Errors.ACCESS_TOKEN_INVALID)


def validate_id_token(
    id_token: str,
    issuer: str,
    client_id: str,
    nonce: str | None,
    verifier: IdTokenVerifier.IdTokenVerifier,
) -> str:
    id_token_components = id_token.split(".")
    if len(id_token_components) != 3:
        raise Exception(Errors.ID_TOKEN_MALFORMED)
    try:
        # Avoid padding errors by appending the max possible padding.
        # b64decode will ignore any extra padding.
        header = json.loads(unquote(b64decode(id_token_components[0] + "==")))
        payload = json.loads(unquote(b64decode(id_token_components[1] + "==")))
    except:
        raise Exception(Errors.ID_TOKEN_HEADER_PAYLOAD_MALFORMED)
    validate_id_token_header(id_token_header=header)
    validate_id_token_payload(
        id_token_payload=payload, issuer=issuer, client_id=client_id, nonce=nonce
    )
    verifier.verify_jwt(id_token)
    return payload["sub"]


def validate_id_token_header(id_token_header: dict) -> None:
    received_alg = id_token_header.get("alg", None)
    if received_alg != "RS256":
        raise Exception(
            get_expected_vs_received_error_message(
                message=Errors.ID_TOKEN_WRONG_SIGNING_ALG,
                expected="RS256",
                received=received_alg,
            )
        )


def validate_id_token_payload(
    id_token_payload: dict,
    issuer: str,
    client_id: str,
    nonce: str | None = None,
) -> None:
    # Check that all required keys are present
    required_keys = ["iss", "sub", "aud", "exp", "iat"]
    present_keys = []
    for required_key in required_keys:
        if required_key in id_token_payload:
            present_keys.append(required_key)
    if len(present_keys) != len(required_keys):
        raise Exception(
            get_expected_vs_received_error_message(
                message=Errors.ID_TOKEN_MISSING_KEYS,
                expected=",".join(required_keys),
                received=",".join(present_keys),
            )
        )

    if id_token_payload["iss"] != issuer:
        raise Exception(
            get_expected_vs_received_error_message(
                message=Errors.ID_TOKEN_ISS_MISMATCH,
                expected=issuer,
                received=id_token_payload["iss"],
            )
        )

    if type(id_token_payload["iat"]) is not int:
        raise Exception(
            get_expected_vs_received_error_message(
                message=Errors.ID_TOKEN_IAT_INVALID,
                expected="a valid number",
                received=id_token_payload["iat"],
            )
        )

    if type(id_token_payload["exp"]) is not int:
        raise Exception(
            get_expected_vs_received_error_message(
                message=Errors.ID_TOKEN_EXP_INVALID,
                expected="a valid number",
                received=id_token_payload["exp"],
            )
        )

    curr_time = datetime.now()
    if id_token_payload["exp"] < curr_time.timestamp():
        raise Exception(
            get_expiry_error_message(
                message=Errors.ID_TOKEN_EXPIRED, expired_at=id_token_payload["exp"]
            )
        )

    if id_token_payload["aud"] != client_id:
        raise Exception(
            get_expected_vs_received_error_message(
                message=Errors.ID_TOKEN_AUD_MISMATCH,
                expected=client_id,
                received=id_token_payload["aud"],
            )
        )

    nonce_received = id_token_payload.get("nonce", None)
    if nonce is not None and nonce_received != nonce:
        raise Exception(
            get_expected_vs_received_error_message(
                message=Errors.ID_TOKEN_NONCE_MISMATCH,
                expected=nonce,
                received=nonce_received,
            )
        )

    if type(id_token_payload["sub"]) is not str or id_token_payload["sub"] == "":
        raise Exception(
            get_expected_vs_received_error_message(
                message=Errors.ID_TOKEN_SUB_INVALID,
                expected="a non-empty string",
                received=id_token_payload["sub"],
            )
        )



from typing import NamedTuple
from .error import Errors
import secrets
from base64 import urlsafe_b64encode
import hashlib


class GeneratePkcePairReturn(NamedTuple):
    code_verifier: str
    code_challenge: str


def generate_code_verifier(length=43) -> str:
    """Generates the random code verifier.

    Args:
        length (int, optional): The length of the code verifier to generate.
        Defaults to 43.

    Raises:
        Exception: if length is <43 or >128.

    Returns:
        str: The generated code verifier.
    """
    if length < 43 or length > 128:
        raise Exception(Errors.CODE_VERIFIER_LENGTH_ERROR)
    bytes = secrets.token_bytes(96)
    encoded = urlsafe_b64encode(bytes)
    return encoded.decode("ascii")[:length]


def generate_code_challenge(code_verifier: str) -> str:
    """Calculates the S256 code challenge for a provided code verifier.

    Args:
        code_verifier (str): The code verifier.

    Returns:
        str: The calculated code challenge.
    """
    verifier_bytearray = bytearray(code_verifier, encoding="ascii")
    code_challenge_hash = hashlib.sha256(verifier_bytearray).digest()
    with_padding = urlsafe_b64encode(code_challenge_hash).decode("ascii")
    # Slice off padding characters as they are not valid for code challenge
    # or verifier. This does not affect the validity of the code challenge
    return with_padding.replace("=", "")


def generate_pkce_pair(length=43) -> GeneratePkcePairReturn:
    """Generates a challenge pair where `code_challenge`
    is the generated S256 hash from `code_verifier`.

    Args:
        length (int, optional): The length of the code verifier. Defaults to 43.

    Raises:
        Exception: if length is <43 or >128.

    Returns:
        GeneratePkcePairReturn: Code challenge and code verifier.
    """
    if length < 43 or length > 128:
        raise Exception(Errors.PKCE_PAIR_LENGTH_ERROR)
    verifier = generate_code_verifier(length)
    challenge = generate_code_challenge(verifier)
    return GeneratePkcePairReturn(code_verifier=verifier, code_challenge=challenge)
